from copy import deepcopy
from enum import IntEnum
from pathlib import Path
from typing import NamedTuple

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QBrush, QColor, QFont
from PyQt6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QFrame, QCheckBox, QSpinBox, \
    QLabel, QListWidget, QListWidgetItem, QButtonGroup, QRadioButton, QComboBox, QSizePolicy
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.step_domain.TemplateDomain import TemplateDomain
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.TemplateSaveFilePickerDialog import \
    TemplateSaveFilePickerDialog
from otlmow_gui.GUI.screens.TemplateScreen_elements.ClassListWidget import ClassListWidget
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.dialog_windows.ChangeSubsetWindow import ChangeSubsetWindow
from otlmow_gui.GUI.screens.screen_interface.TemplateScreenInterface import TemplateScreenInterface
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class TemplateScreen(TemplateScreenInterface):
    """Represents the template screen for managing project settings and data export.

    This class provides a user interface for exporting templates.
    It includes various UI components such as labels, buttons, checkboxes, and lists to facilitate
    user interactions and display relevant information.

    Args:f
        language_settings (optional): Language settings for the user interface.

    Attributes:
        project (optional): The current project associated with the template screen.
        main_window (optional): Reference to the main application window.
        main_layout (QVBoxLayout): Layout for organizing UI components vertically.
        subset_name (QLabel): Label for displaying the subset name.
        operator_name (QLabel): Label for displaying the operator's name.
        otl_version (QLabel): Label for displaying the OTL version.
        operator_title (QLabel): Label for the operator title.
        otl_title (QLabel): Label for the OTL title.
        change_subset_btn (ButtonWidget): Button for changing the subset.
        general_settings_title (QLabel): Title label for general settings.
        add_choice_list (QCheckBox): Checkbox for no choice list option.
        add_geometry_attributes (QCheckBox): Checkbox for geometry column addition.
        export_button (ButtonWidget): Button for exporting data.
        export_attribute_info (QCheckBox): Checkbox for exporting attribute information.
        show_deprecated_attributes (QCheckBox): Checkbox for showing deprecated attributes.
        example_amount_checkbox (QLabel): Label for the example amount setting.
        non_otl_conform_settings_title (QLabel): Title label for non-OTL conform settings.
        amount_of_examples (QSpinBox): Spin box for specifying the amount of examples.
        example_settings_title (QLabel): Title label for example settings.
        select_all_classes (QCheckBox): Checkbox for selecting all classes.
        all_classes (QListWidget): List widget for displaying available classes.
        selected (int): Counter for the number of selected classes.
        label_counter (QLabel): Label for displaying the count of selected classes.
    """

    class TemplateOptionId(IntEnum):
        DAVIE_CONFORM = 0
        EXTRA_INFO = 1




    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings

        self.project = None

        self.main_window = None
        self.main_layout = QVBoxLayout()

        # project information overview box
        self.subset_name = QLabel()
        self.operator_name = QLabel()
        self.otl_version = QLabel()
        self.operator_title = QLabel()
        self.otl_title = QLabel()
        self.change_subset_btn = ButtonWidget()

        # Combobox to select the export file type
        self.file_extension_selection: QComboBox = QComboBox()
        self.add_resetable_field(self.file_extension_selection)
        self.file_type_label: QLabel = QLabel()
        self.supported_export_formats: dict = deepcopy(global_vars.supported_file_formats)

        # if "SDF" in self.supported_export_formats:
        #     self.supported_export_formats.pop("SDF")  # not yet supported for export in V0.5.3
        if "JSON" in self.supported_export_formats:
            self.supported_export_formats.pop("JSON")  # Doesn't support template creation
        if "GeoJSON" in self.supported_export_formats:
            self.supported_export_formats.pop("GeoJSON")  # Doesn't support template creation

        FileTypeSettingPropertySetting = NamedTuple("FileTypeSettingPropertySetting",[
            ("enabled",bool),
            ("change_state_if_enabled",bool),
            ("default_on",bool),
            ("tooltip",str)])
        FileTypeSetting = NamedTuple("FileTypeSetting", [
            ("choice_lists",FileTypeSettingPropertySetting),
            ("example_assets",FileTypeSettingPropertySetting),
            ("geometry_attributes", FileTypeSettingPropertySetting),
            ("expansive_info", FileTypeSettingPropertySetting)])

        self.default_choice_list_tooltip = self._("Adds choice lists to relevant "
                                                           "attributes")
        self.always_on_choice_list_tooltip = self._(
            "For {filetype} choice lists cannot be disabled")
        self.always_off_choice_list_tooltip = self._("{filetype} cannot contain choice lists")

        self.default_geometry_tooltip = self._("Adds geometry information for each asset")
        self.always_on_geometry_tooltip = self._(
            "For {filetype} geometry attributes cannot be disabled")
        self.always_off_geometry_tooltip = self._("{filetype} cannot contain geometry attributes")

        self.default_example_checkbox_tooltip = self._("Adds example assets to each OTL-class")
        self.always_on_example_checkbox_tooltip = self._(
            "For {filetype} generated examples cannot be disabled")
        self.always_off_example_checkbox_tooltip = self._(
            "{filetype} cannot add generated examples")

        self.default_expansive_info_radiobutton_tooltip = (
            self._("Adds extra rows with a description of every attribute and, an indication if the"
                   " attribute is deprecated \n(Delete extra rows before uploading to DAVIE)"))
        self.always_on_expansive_info_radiobutton_tooltip = self._(
            "For {filetype} expansive information cannot be disabled")
        self.always_off_expansive_info_radiobutton_tooltip = self._(
            "{filetype} cannot add expansive information")

        self.file_type_settings:dict[str,FileTypeSetting] = {}
        self.file_type_settings = {"Excel":FileTypeSetting(
                                        choice_lists= FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip= self.default_choice_list_tooltip.format(filetype="Excel")
                                        ),
                                        geometry_attributes= FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip=self.default_geometry_tooltip
                                        ),
                                        example_assets= FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=True,
                                            default_on=False,
                                            tooltip=self.default_example_checkbox_tooltip
                                        ),
                                        expansive_info=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.default_expansive_info_radiobutton_tooltip
                                        ),
                                    ),
                                    "CSV": FileTypeSetting(
                                        choice_lists=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=True,
                                            default_on=False,
                                            tooltip=self.always_off_choice_list_tooltip.format(filetype="CSV")
                                        ),
                                        geometry_attributes=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip=self.default_geometry_tooltip
                                        ),
                                        example_assets=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=True,
                                            tooltip=self.default_example_checkbox_tooltip
                                        ),
                                        expansive_info=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.default_expansive_info_radiobutton_tooltip
                                        ),
                                    ),
                                    "JSON": FileTypeSetting(
                                        choice_lists=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=True,
                                            default_on=False,
                                            tooltip=self.always_off_choice_list_tooltip.format(filetype="JSON")
                                        ),
                                        geometry_attributes=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip=self.default_geometry_tooltip
                                        ),
                                        example_assets=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.default_example_checkbox_tooltip
                                        ),
                                        expansive_info=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.default_expansive_info_radiobutton_tooltip
                                        ),
                                    ),
                                    'GeoJSON': FileTypeSetting(
                                        choice_lists=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=True,
                                            default_on=False,
                                            tooltip=self.always_off_choice_list_tooltip.format(filetype="GeoJSON")
                                        ),
                                        geometry_attributes=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip=self.default_geometry_tooltip
                                        ),
                                        example_assets=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.default_example_checkbox_tooltip
                                        ),
                                        expansive_info=FileTypeSettingPropertySetting(
                                            enabled=True,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.default_expansive_info_radiobutton_tooltip
                                        ),
                                    ),
                                    'SDF': FileTypeSetting(
                                        choice_lists=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip=self.always_on_choice_list_tooltip.format(filetype="SDF")
                                        ),
                                        geometry_attributes=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=True,
                                            default_on=True,
                                            tooltip=self.always_on_geometry_tooltip.format(filetype="SDF")
                                        ),
                                        example_assets=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.always_off_example_checkbox_tooltip.format(filetype="SDF")
                                        ),
                                        expansive_info=FileTypeSettingPropertySetting(
                                            enabled=False,
                                            change_state_if_enabled=False,
                                            default_on=False,
                                            tooltip=self.always_off_expansive_info_radiobutton_tooltip.format(filetype="SDF")
                                        ),
                                    ),
                                    }

        self.template_export_file_picker = TemplateSaveFilePickerDialog(self._)

        self.has_agent = False
        # settings checkboxes and counters
        self.general_settings_title = QLabel()
        self.add_choice_list = QCheckBox()
        self.add_geometry_attributes = QCheckBox()
        self.export_button = ButtonWidget()
        self.export_attribute_info = QCheckBox()
        self.show_deprecated_attributes = QCheckBox()
        self.example_amount_checkbox = QCheckBox()
        self.non_otl_conform_settings_title = QLabel()
        self.amount_of_examples = QSpinBox()
        self.example_settings_title = QLabel()

        # class GUI list elements
        self.select_all_classes = QCheckBox()
        self.add_resetable_field(self.select_all_classes,default_value=True)
        self.all_classes = ClassListWidget()
        self.selected = 0
        self.label_counter = QLabel()

        # radio button for DAVIE conformity or extra info
        self.template_format_label = QLabel()
        self.radio_button_group = QButtonGroup()
        self.radio_button_davie_conform = QRadioButton()
        self.add_resetable_field(self.radio_button_davie_conform,True)
        self.radio_button_expanded_info = QRadioButton()
        self.add_resetable_field(self.radio_button_expanded_info, False)


        self.init_ui()
        self.update_settings_based_on_filetype(self.file_extension_selection.currentText())

    def init_ui(self) -> None:
        """
        Initializes the user interface layout for the template screen.

        This function sets up the main layout by adding spacing, a template menu widget
         and adjusting the layout's margins.
         It ensures that the user interface is properly structured and visually appealing.

        :param self: The instance of the class.
        :return: None
        """

        self.main_layout.addSpacing(10)
        self.main_layout.addWidget(self.template_menu())
        # self.main_layout.addStretch()
        self.main_layout.addSpacing(10)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.main_layout)


    def options_menu(self) -> QFrame:
        """
        Creates and configures the options menu for the user interface.

        This function builds the options menu by adding various settings and controls,
        such as checkboxes and buttons, to the layout. It sets properties and connects signals
        to ensure that user interactions are handled appropriately.

        :param self: The instance of the class.
        :return: QFrame -- The configured options menu as a QFrame widget.
        """

        options_menu = QFrame()
        main_layout = QVBoxLayout()


        self.general_settings_title.setProperty('class', 'settings-title')
        self.general_settings_title.setText(self._("general_settings"))

        self.add_choice_list.setText(self._("choice_list"))
        self.add_choice_list.setProperty('class', 'settings-checkbox')

        self.add_geometry_attributes.setText(self._("geometry_attributes"))
        self.add_geometry_attributes.setProperty('class', 'settings-checkbox')
        self.add_geometry_attributes.setChecked(True)

        self.non_otl_conform_settings_title.setText(self._("add_non_otl_conform_information"))
        self.non_otl_conform_settings_title.setProperty('class', 'settings-title')

        self.show_deprecated_attributes.setText(self._("show_deprecated_attributes"))
        self.show_deprecated_attributes.setProperty('class', 'settings-checkbox')
        self.show_deprecated_attributes.setEnabled(False)

        self.export_attribute_info.setText(self._("export_attribute_info"))
        self.export_attribute_info.setProperty('class', 'settings-checkbox')

        self.example_settings_title.setProperty('class', 'settings-title')
        self.example_settings_title.setText(self._("example_settings"))

        self.export_button.setText(self._("export"))
        self.export_button.setProperty('class', 'primary-button')
        self.export_button.clicked.connect(lambda: self.export_template_listener())

        # build main layout
        main_layout.addWidget(self.create_filetype_combobox())
        main_layout.addWidget(self.create_radio_button_box())
        main_layout.addSpacing(10)

        main_layout.addWidget(self.add_choice_list)
        main_layout.addWidget(self.add_geometry_attributes)
        main_layout.addWidget(self.create_example_generation_container())
        main_layout.addSpacing(10)
        main_layout.addWidget(self.export_button, alignment=Qt.AlignmentFlag.AlignLeft)

        main_layout.addStretch()

        options_menu.setLayout(main_layout)

        return options_menu

    def create_radio_button_box(self) -> QFrame:


        button_box_frame = QFrame()
        button_box_layout = QHBoxLayout()

        # self.template_format_label.setProperty('class', 'settings-checkbox')
        self.template_format_label.setText(self._("Template format") + ":")
        self.template_format_label.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self.radio_button_davie_conform.setText(self._('DAVIE conform'))
        self.radio_button_davie_conform.setToolTip(self._("Generates a template that is ready for upload to DAVIE"))

        self.radio_button_expanded_info.setText(self._('Expanded info'))
        self.radio_button_expanded_info.setToolTip(self.default_expansive_info_radiobutton_tooltip)

        self.radio_button_group.addButton(self.radio_button_davie_conform, self.TemplateOptionId.DAVIE_CONFORM)
        # self.button_group.setId(self.radio_button_export_all_data,
        #                         self.ExportOptionId.ALL_DATA)

        self.radio_button_group.addButton(self.radio_button_expanded_info, self.TemplateOptionId.EXTRA_INFO)

        self.radio_button_davie_conform.setChecked(True)

        button_box_layout.addWidget(self.template_format_label)
        button_box_layout.addStretch()
        button_box_layout.addWidget(self.radio_button_davie_conform)
        button_box_layout.addStretch()
        button_box_layout.addWidget(self.radio_button_expanded_info)
        button_box_layout.addStretch()

        button_box_layout.setContentsMargins(0,0,0,0)
        button_box_frame.setLayout(button_box_layout)


        return button_box_frame

    # def on_radio_button_click(self,id:int):
    #     if id == self.TemplateOptionId.DAVIE_CONFORM:
    #         OTLLogger.logger.debug("Generate DAVIE conform templates")
    #     elif id == self.TemplateOptionId.ONLY_UNEDITED_DATA:
    #         OTLLogger.logger.debug("Generate extra information templates")

    def create_example_generation_container(self) -> QFrame:
        """
        Creates a container for example generation settings.

        This function initializes a QFrame that contains a label and a spin box for specifying the
        amount of examples to generate. It sets the appropriate properties and layout to ensure a
        user-friendly interface.

        :param self: The instance of the class.
        :return: QFrame -- The configured example generation container.
        """

        example_generation_container = QFrame()
        example_box_layout = QHBoxLayout()
        self.example_amount_checkbox.setText(self._("amount_of_examples"))
        self.example_amount_checkbox.setToolTip(self.default_example_checkbox_tooltip)
        # self.example_amount_label.setProperty('class', 'settings-label')
        self.amount_of_examples.setRange(1, 100)
        self.amount_of_examples.setValue(1)
        self.amount_of_examples.setToolTip(self._("The amount of example assets added to each OTL-class"))
        example_box_layout.addWidget(self.example_amount_checkbox)
        example_box_layout.addWidget(self.amount_of_examples)
        example_box_layout.setContentsMargins(0,0,0,0)

        example_generation_container.setLayout(example_box_layout)
        example_generation_container.setProperty('class', 'settings-checkbox')

        return example_generation_container


    def template_menu(self) -> QWidget:
        """
        Constructs the template menu for the user interface.

        This function creates a QWidget that serves as the main container for the template menu,
        including a title, options menu, and a list. It organizes these components using horizontal
        and vertical layouts to ensure a structured and visually appealing interface.

        :param self: The instance of the class.
        :return: QWidget -- The constructed template menu as a QWidget.
        """

        full_window = QWidget()
        full_window.setProperty('class', 'background-box')
        horizontal_layout = QHBoxLayout()
        window = QFrame()
        layout = QVBoxLayout()
        layout.addWidget(self.subset_title_and_button())
        layout.addWidget(self.options_menu())
        layout.setContentsMargins(16, 0, 16, 0)
        window.setLayout(layout)
        horizontal_layout.addWidget(window,stretch=1)
        horizontal_layout.addWidget(self.create_list(),stretch=2)
        horizontal_layout.addSpacing(20)
        full_window.setLayout(horizontal_layout)


        return full_window


    def subset_title_and_button(self) -> QFrame:
        """
        Creates a frame containing subset information and a button.

        This function constructs a QFrame that includes a list of subset information and a button
        to change the subset. It organizes these components in a horizontal layout, ensuring that
        the button is aligned to the right for a clean user interface.

        :param self: The instance of the class.
        :return: QFrame -- The constructed frame with subset information and button.
        """

        frame = QFrame()
        horizontal_layout = QHBoxLayout()
        self.change_subset_btn.setText(self._("change_subset"))
        self.change_subset_btn.clicked.connect(lambda: self.change_subset())
        self.change_subset_btn.setProperty('class', 'secondary-button')
        horizontal_layout.addWidget(self.subset_info_list())
        horizontal_layout.addWidget(self.change_subset_btn, alignment=Qt.AlignmentFlag.AlignRight)
        horizontal_layout.setContentsMargins(0, 16, 0, 16)
        frame.setLayout(horizontal_layout)
        return frame


    def create_operator_info_field(self) -> QFrame:
        """
        Constructs a field displaying operator information.

        This function creates a QFrame that contains a label for the operator and an input field
        for the operator's name. It organizes these components in a horizontal layout, ensuring
        that the label is clearly associated with the input field.

        :param self: The instance of the class.
        :return: QFrame -- The constructed frame with operator information.
        """

        frame = QFrame()
        horizontal_layout = QHBoxLayout()
        self.operator_title.setText(self._("operator") + ":")
        self.operator_title.setProperty('class', 'info-label')
        self.operator_name.setText("")
        horizontal_layout.addWidget(self.operator_title)
        horizontal_layout.addWidget(self.operator_name, alignment=Qt.AlignmentFlag.AlignRight)
        horizontal_layout.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(horizontal_layout)
        return frame


    def create_otl_version_field(self) -> QFrame:
        """
        Constructs a field for displaying the OTL version information.

        This function creates a QFrame that contains a label for the OTL version and an input field
        for displaying the version. It arranges these components in a horizontal layout, ensuring
        that the label is clearly associated with the version input field.

        :param self: The instance of the class.
        :return: QFrame -- The constructed frame with OTL version information.
        """

        frame = QFrame()
        horizontal_layout = QHBoxLayout()
        self.otl_title.setText(self._("otl_version") + ":")
        self.otl_title.setProperty('class', 'info-label')
        self.otl_version.setText("")
        horizontal_layout.addWidget(self.otl_title)
        horizontal_layout.addWidget(self.otl_version, alignment=Qt.AlignmentFlag.AlignRight)
        horizontal_layout.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(horizontal_layout)
        return frame


    def create_subset_name_field(self) -> QFrame:
        """
        Constructs a field for entering the subset name.

        This function creates a QFrame that includes a label for the subset and an input field for
        entering the subset name. It organizes these components in a horizontal layout, ensuring
        that the label is clearly associated with the input field.

        :param self: The instance of the class.
        :return: QFrame -- The constructed frame for the subset name input.
        """

        frame = QFrame()
        horizontal_layout = QHBoxLayout()
        subset_title = QLabel()
        subset_title.setText(self._("subset") + ":")
        subset_title.setProperty('class', 'info-label')
        self.subset_name.setText("")
        horizontal_layout.addWidget(subset_title)
        horizontal_layout.addWidget(self.subset_name, alignment=Qt.AlignmentFlag.AlignRight)
        horizontal_layout.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(horizontal_layout)
        return frame


    def subset_info_list(self) -> QFrame:
        """Constructs a frame containing information fields for subsets.

        This function creates a QFrame that includes various input fields related to subsets,
        such as the subset name, operator information, and OTL version. It organizes these
        components in a vertical layout to provide a clear and structured user interface.

        :param self: The instance of the class.
        :return: QFrame -- The constructed frame containing subset information fields.
        """

        frame = QFrame()
        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(self.create_subset_name_field())
        vertical_layout.addWidget(self.create_operator_info_field())
        vertical_layout.addWidget(self.create_otl_version_field())
        frame.setLayout(vertical_layout)
        return frame


    def create_list(self) -> QFrame:
        """
        Constructs a list interface for selecting classes.

        This function creates a QFrame that contains a checkbox for selecting all classes, a list
        widget for displaying available classes, and a label that shows the count of selected
        classes. It sets up the necessary connections to handle user interactions and updates the
        display accordingly.

        :param self: The instance of the class.
        :return: QFrame -- The constructed frame containing the class selection interface.
        """

        frame = QFrame()
        vertical_layout = QVBoxLayout()

        self.select_all_classes.setText(self._("select_all_classes"))
        self.select_all_classes.clicked.connect(lambda: self.select_all_classes_clicked())
        self.all_classes.setSelectionMode(QListWidget.SelectionMode.MultiSelection)

        self.all_classes.doubleClicked.connect(
            lambda item:
                 self.all_classes.itemFromIndex(item).setSelected(not self.all_classes.itemFromIndex(item).isSelected())
        )
        self.label_counter.setText(self._(f"{self.selected} classes selected"))

        vertical_layout.addWidget(self.select_all_classes, alignment=Qt.AlignmentFlag.AlignTop)
        vertical_layout.addWidget(self.all_classes)
        vertical_layout.addWidget(self.label_counter)
        frame.setLayout(vertical_layout)

        return frame


    def update_project_info(self, project: Project) -> None:
        """Updates the project information displayed in the user interface.

        This function retrieves the current subset name, operator name, and OTL version from the
        project and updates the corresponding UI fields. If the project data cannot be found,
        it sets the fields to default values to indicate the absence of information.

        :param self: The instance of the class.
        :param project: the new project to which the information needs to be updated to
        :return: None
        :raises FileNotFoundError: If the project data cannot be accessed.
        """
        if project:
            try:
                # model_builder = ModelBuilder(self.project.subset_path)
                self.subset_name.setText(project.get_subset_db_name())
                self.operator_name.setText(project.get_operator_name())
                self.otl_version.setText(project.get_otl_version())
            except FileNotFoundError as e:
                self.subset_name.setText("/")
                self.operator_name.setText("/")
                self.otl_version.setText("/")
        else:
            self.subset_name.setText("Loading")
            self.operator_name.setText("Loading")
            self.otl_version.setText("Loading")

    def class_items_clicked_listener(self, selected_item):
        row_index = selected_item.data(1)[1]
        TemplateDomain.toggle_class_index(row_index)




    def update_label_under_list(self,total_amount_of_items:int,counter:int) -> None:

        self.selected = counter
        self.label_counter.setText(self._("{selected} classes selected").format(selected=self.selected))
        if counter:
            self.export_button.setEnabled(True)
            self.export_button.setToolTip(self._("Export template"))
        else:
            self.export_button.setEnabled(False)
            self.export_button.setToolTip(self._("Select at least 1 class to export"))

    def update_all_classes_selected(self,state:bool) -> None:
        self.select_all_classes.setChecked(state)

    def count_selected_items(self):
        return sum(
            1
            for i in range(self.all_classes.count())
            if self.all_classes.item(i).isSelected()
        )



    def select_all_classes_clicked(self):
        """
        Handles the selection of all classes based on the checkbox state.

        This function checks if the "select all classes" checkbox is checked and, if so, selects 
        all items in the class list. If the checkbox is unchecked, it clears the selection. 
        The function does nothing if the class list is disabled.

        :param self: The instance of the class.
        :return: None
        """

        if not self.all_classes.isEnabled():
            return
        # elif self.select_all_classes.isChecked():
        #     TemplateDomain.select_all_classes()
        #
        #     # self.all_classes.selectAll()
        # else:
        #     TemplateDomain.deselect_all_classes()

        TemplateDomain.set_select_all_classes(self.select_all_classes.isChecked())

    def deselect_all_classes(self) -> None:
        self.all_classes.clearSelection()


    def set_all_classes_selected(self) -> None:
        self.all_classes.selectAll()

    def export_template_listener(self) -> None:  # sourcery skip: use-named-expression
        """
        Handles the export of a template based on selected classes.
        Handles the export of a template based on selected classes.

        This function retrieves the selected classes from the user interface and prompts the user 
        to choose a file location for the export. If a valid file path is provided, it calls the 
        export function with the necessary parameters, including user preferences for the 
        export options.

        :param self: The instance of the class.
        :return: None
        """

        chosen_file_format = self.file_extension_selection.currentText()
        if chosen_file_format in self.supported_export_formats:
            selection_path_list = self.template_export_file_picker.summon(
                chosen_file_format=chosen_file_format,
                supported_export_formats=self.supported_export_formats,
                project_name=global_vars.current_project.eigen_referentie)
            try:
                self.save_template(selection_path_list)
            except Exception as e:
                # TODO: proper error message when template export fails
                raise e

    def save_template(self, document_paths:list[Path]):

        if not document_paths or not document_paths[0]:
            return
        document_path = document_paths[0]

        checked_radio_button_id = self.radio_button_group.checkedId()
        if checked_radio_button_id == self.TemplateOptionId.DAVIE_CONFORM:
            attribute_description = False
            deprecated_attribute_marker = False
        else:
            attribute_description = True
            deprecated_attribute_marker = True

        if self.example_amount_checkbox.isChecked():
            amount_of_examples = self.amount_of_examples.value()
        else:
            amount_of_examples = 0

        create_task_reraise_exception(TemplateDomain.async_export_template(
            document_path=document_path,
            generate_choice_list=self.add_choice_list.isChecked(),
            geometry_column_added=self.add_geometry_attributes.isChecked(),
            export_attribute_info=attribute_description,
            highlight_deprecated_attributes=deprecated_attribute_marker,
            amount_of_examples=amount_of_examples))


    def update_settings_based_on_filetype(self,filetype:str):
        # if filetype == "Excel":

        self.add_choice_list.setChecked(self.file_type_settings[filetype].choice_lists.default_on)
        self.add_choice_list.setEnabled(self.file_type_settings[filetype].choice_lists.enabled)
        self.add_choice_list.setToolTip(self.file_type_settings[filetype].choice_lists.tooltip)

        self.example_amount_checkbox.setChecked(self.file_type_settings[filetype].example_assets.default_on)
        self.example_amount_checkbox.setEnabled(self.file_type_settings[filetype].example_assets.enabled)
        self.example_amount_checkbox.setToolTip(self.file_type_settings[filetype].example_assets.tooltip)

        self.add_geometry_attributes.setChecked(self.file_type_settings[filetype].geometry_attributes.default_on)
        self.add_geometry_attributes.setEnabled(self.file_type_settings[filetype].geometry_attributes.enabled)
        self.add_geometry_attributes.setToolTip(self.file_type_settings[filetype].geometry_attributes.tooltip)

        if ((not self.file_type_settings[filetype].expansive_info.enabled) and
            self.radio_button_expanded_info.isChecked()):
            self.radio_button_expanded_info.setChecked(False)
            self.radio_button_davie_conform.setChecked(True)

        self.radio_button_expanded_info.setEnabled(self.file_type_settings[filetype].expansive_info.enabled)
        self.radio_button_expanded_info.setToolTip(self.file_type_settings[filetype].expansive_info.tooltip)


            # self.add_choice_list.setChecked(True)
            # self.add_choice_list.setEnabled(True)
            # self.add_choice_list.setToolTip(
            #     self._("Adds choice lists to relevant attributes").format(filetype=filetype))
            #
            # self.add_geometry_attributes.setChecked(True)
            # self.add_geometry_attributes.setEnabled(True)
            # self.add_geometry_attributes.setToolTip(
            #     self._("Adds geometry information for each asset"))
        # elif filetype == 'CSV':
        #     self.add_choice_list.setChecked(False)
        #     self.add_choice_list.setEnabled(False)
        #     self.add_choice_list.setToolTip(self._("{filetype} cannot contain choice lists").format(filetype=filetype))
        #
        #     self.add_geometry_attributes.setChecked(True)
        #     self.add_geometry_attributes.setEnabled(True)
        #     self.add_geometry_attributes.setToolTip(
        #         self._("Adds geometry information for each asset"))
        # elif filetype == 'JSON':
        #     self.add_choice_list.setChecked(False)
        #     self.add_choice_list.setEnabled(False)
        #     self.add_choice_list.setToolTip(self._("{filetype} cannot contain choice lists").format(filetype=filetype))
        #
        #     self.add_geometry_attributes.setChecked(True)
        #     self.add_geometry_attributes.setEnabled(True)
        #     self.add_geometry_attributes.setToolTip(
        #         self._("Adds geometry information for each asset"))
        # elif filetype == 'GeoJSON':
        #     self.add_choice_list.setChecked(False)
        #     self.add_choice_list.setEnabled(False)
        #     self.add_choice_list.setToolTip(self._("{filetype} cannot contain choice lists").format(filetype=filetype))
        #
        #     self.add_geometry_attributes.setChecked(True)
        # elif filetype == 'SDF':
        #     self.add_choice_list.setChecked(True)
        #     self.add_choice_list.setEnabled(False)
        #     self.add_choice_list.setToolTip(self._("For {filetype} choice lists cannot be disabled").format(filetype=filetype))
        #
        #     self.add_geometry_attributes.setChecked(True)
        #     self.add_geometry_attributes.setEnabled(False)
        #     self.add_choice_list.setToolTip(self._("For {filetype} geometry attributes cannot be disabled").format(filetype=filetype))
    def change_subset(self) -> None:
        """
        Opens a window to change the current subset.

        This function initializes and displays the ChangeSubsetWindow, allowing the user to modify 
        the current subset settings. It provides a user interface for selecting and applying 
        changes to the subset.

        :param self: The instance of the class.
        :return: None
        """

        ChangeSubsetWindow(self._)


    def reset_ui(self, _) -> None:
        super().reset_ui(_)
        self.update_settings_based_on_filetype(self.file_extension_selection.currentText())
        self._ = _

            
        # self.export_attribute_info.setText(self._("export_attribute_info"))
        # self.add_geometry_attributes.setText(self._("geometry_column_added"))
        # self.add_choice_list.setText(self._("choice_list"))
        # self.select_all_classes.setText(self._("select_all_classes"))
        # self.example_amount_checkbox.setText(self._("amount_of_examples"))
        # self.export_button.setText(self._("export"))
        # self.change_subset_btn.setText(self._("change_subset"))
        # self.operator_title.setText(self._("operator") + ":")
        # self.otl_title.setText(self._("otl_version") + ":")
        # self.general_settings_title.setText(self._("general_settings"))
        # self.example_settings_title.setText(self._("example_settings"))
        # self.non_otl_conform_settings_title.setText(self._("deprecated_settings"))
        # self.label_counter.setText(self._("{selected} classes selected").format(selected=self.selected))


    def set_gui_list_to_loading_state(self) -> None:
        """Sets the GUI list to a loading state.

        This function updates the list of classes in the user interface to indicate that loading is in progress. It clears the current items in the list and adds a single item displaying the text "loading" to inform the user of the ongoing operation.

        :param self: The instance of the class.
        :return: None
        """

        item = QListWidgetItem()
        item.setText(self._("loading"))
        self.all_classes.clear()
        self.all_classes.addItem(item)


    def set_gui_list_to_no_classes_found(self) -> None:
        """
        Updates the GUI list to indicate that no classes were found.

        This function disables the class list in the user interface and adds a message indicating
        that no classes were found in the specified path. It provides feedback to the user when
        the expected data is not available.

        :param self: The instance of the class.
        :return: None
        """

        self.all_classes.setEnabled(False)
        self.all_classes.addItem(self._("no classes found in specified path"))


    def set_classes(self, classes: list[OSLOClass],
                    selected_classes: list[int],
                    all_classes_selected_checked:bool,
                    has_a_class_with_deprecated_attributes: bool) -> None:
        """
        Sets the list of classes in the user interface.

        This function clears the existing items in the class list and populates it with new items
        based on the provided list of OSLOClass instances. It also enables or disables the display
        of deprecated attributes based on the provided flag.

        :param self: The instance of the class.
        :param classes: A list of OSLOClass instances to be displayed in the class list.
        :type classes: list[OSLOClass]
        :param has_a_class_with_deprecated_attributes: A flag indicating whether to show deprecated attributes.
        :type has_a_class_with_deprecated_attributes: bool
        :return: None
        """

        self.all_classes.clear()
        self.all_classes.setEnabled(True)
        self.has_agent = False

        # store indexes for class list in backend before sorting!
        class_index_tuple_list = [(OTL_class,i) for i,OTL_class in enumerate(classes)]
        sorted_classes = sorted(class_index_tuple_list,key=lambda x : x[0].name)

        for value_tuple in sorted_classes:
            class_name = value_tuple[0].name
            class_typeURI = value_tuple[0].objectUri
            backend_index = value_tuple[1]

            item = QListWidgetItem()
            item.setText(class_name)
            item.setData(1, (class_typeURI,backend_index))



            if class_name == "Agent":
                item.setBackground(QBrush(QColor("#AAAAAA")))
                item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEnabled)
                placeholder_font = QFont()
                placeholder_font.setItalic(True)
                item.setFont(placeholder_font)
                item.setToolTip(self._('Agent class is not allowed in the template'))
                self.has_agent = True


            self.all_classes.addItem(item)

        if has_a_class_with_deprecated_attributes:
            self.show_deprecated_attributes.setEnabled(False)

        if all_classes_selected_checked:
            self.set_all_classes_selected()
        else:
            self.update_all_classes_selected(False)
            # the need to be added to cls.all_classes before you can select them
            for class_display_index in range(self.all_classes.count()):
                item = self.all_classes.item(class_display_index)
                real_index = item.data(1)[1]

                if real_index in selected_classes:
                    item.setSelected(True)

        self.update_label_under_list(
            total_amount_of_items=TemplateDomain.get_total_amount_of_classes(),
            counter=len(selected_classes))


    def create_filetype_combobox(self) -> QFrame:
        """
        Creates a combo box for selecting the file type for export. This method sets up the layout,
        adds a label and a combo box populated with supported file formats, and connects the c
        ombo box's change event to a method for displaying additional options.

        :return: A QFrame containing the combo box for file type selection.
        :rtype: QFrame
        """

        frame = QFrame()
        frame_layout = QHBoxLayout()

        self.file_type_label.setText(self._('select file type for export') + ":")

        self.file_extension_selection.addItems(list(self.supported_export_formats.keys()))
        self.file_extension_selection.currentTextChanged.connect(self.update_settings_based_on_filetype)
        self.file_extension_selection.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Fixed))

        frame_layout.addWidget(self.file_type_label)
        frame_layout.addWidget(self.file_extension_selection)
        frame_layout.setContentsMargins(0,0,0,0)
        frame.setLayout(frame_layout)
        frame.setToolTip(self._("The type of file the template will be created in"))

        return frame

    def open_folder_of_created_template(self,document_path:Path):
        Helpers.open_folder_and_select_document(document_path=document_path)



import abc
from typing import Optional, Collection

from PyQt6.QtCore import Qt, QModelIndex, QItemSelectionModel
from PyQt6.QtGui import QColor, QStandardItem, QPixmap, QIcon, QPainter, QBrush, QFont
from PyQt6.QtWidgets import QTreeWidget, QFrame, QVBoxLayout, QLabel, QListWidget, QListWidgetItem, \
    QHeaderView, QTreeWidgetItem, QHBoxLayout, QLineEdit, QPushButton, QTreeView, \
    QStyledItemDelegate

import qtawesome as qta
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.AIMObject import AIMObject
from otlmow_visuals.PyVisWrapper import PyVisWrapper

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.Styling import Styling
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.screens.RelationChange_elements.FolderTreeView import FolderTreeView

IMG_DIR = ProgramFileStructure.get_dynamic_library_path('img')
MULTI_SELECTION = QListWidget.SelectionMode.MultiSelection


class AbstractInstanceListWidget:

    def __init__(self, language_settings,parent,labels,list_gui_style_class=None,needs_source_object= False):
        self._ = language_settings
        self.parent = parent
        self.search_text = ""

        self.search_bar = None
        self.clear_search_bar_button = None

        self.list_gui: Optional[FolderTreeView] = None
        self.list_button = ButtonWidget()

        self.list_label = None
        self.list_subtext_frame = None
        self.list_subtext_layout = None
        self.list_subtext_label = None
        self.attribute_field: QTreeWidget = QTreeWidget()
        self.attribute_field.setSelectionMode(QTreeWidget.SelectionMode.NoSelection)
        self.frame_layout = None

        # self.type_to_items_dict = {}
        self.type_open_status = {}
        self.selected_object = None
        self.selected_item = None
        self.list_gui_style_class = list_gui_style_class

        self.item_type_data_index = 3
        self.data_1_index = 4
        self.data_last_added_index = 5
        self.data_item_count_index = 8

        # self.data_1_index = 4
        # self.data_2_index = 5
        # self.data_3_index = 8
        # self.data_last_added_index = 9

        self.color_legend = PyVisWrapper().relatie_color_dict
        self.needs_source_object= needs_source_object

        self.labels = labels
        self.first_paint = False
        self.second_paint = False
        self.multi_col_list = True

        self.id_to_object_with_text_and_data_dict: dict = {}


    class LastAddedHighlightDelegate(QStyledItemDelegate):

        def __init__(self, list_widget):

            super().__init__()
            self.parent : AbstractInstanceListWidget = list_widget
            self.second_paint = False
            self.first_paint = False

        def paint(self, painter: QPainter, option, index: QModelIndex):
            painter.save()

            if(self.first_paint and not self.second_paint):
                self.parent.list_gui.resize_columns(self.parent.multi_col_list)
                self.second_paint = True
            else:
                self.first_paint = True

            # Apply custom background for specific rows or items
            if  self.parent.list_gui.model.itemFromIndex(index.siblingAtColumn(0)).data(self.parent.data_last_added_index) :
                painter.fillRect(option.rect, QBrush(Styling.last_added_color))

            painter.restore()
            super().paint(painter, option, index)

    def create_object_list_gui(self, multi_select: bool = False) -> QFrame:
        frame = QFrame()
        self.frame_layout = QVBoxLayout()
        self.list_label = QLabel()
        self.list_label.setText(self.list_label_text)

        list_label_font = QFont()
        list_label_font.setPointSize(11)
        list_label_font.setBold(True)
        self.list_label.setFont(list_label_font)
        # self.list_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.list_subtext_frame = self.create_list_subtext_frame()

        self.list_gui = FolderTreeView()
        self.list_gui.setProperty('class', 'list')
        self.list_gui.selectionModel().selectionChanged.connect(self.on_item_selectionChange_listener)
        self.list_gui.expanded.connect(self.record_expanse_listener)
        self.list_gui.collapsed.connect(self.record_collapse_listener)
        self.list_gui.clicked.connect(self.clicked_item_listener)
        self.list_gui.setExpandsOnDoubleClick(False)
        self.list_gui.setItemDelegate(self.LastAddedHighlightDelegate(self))
        if multi_select:
            self.list_gui.setSelectionMode(QTreeView.SelectionMode.MultiSelection)
        if self.list_gui_style_class:
            self.list_gui.setObjectName(self.list_gui_style_class)




        self.frame_layout.addWidget(self.list_label)
        self.frame_layout.addWidget(self.list_subtext_frame)

        self.frame_layout.addWidget(self.create_search_bar())
        self.frame_layout.addWidget(self.list_gui,10)

        object_attribute_label = QLabel()
        object_attribute_label.setText(self.attribute_field_label_text)
        self.frame_layout.addWidget(self.create_button())
        self.frame_layout.addWidget(object_attribute_label)

        self.frame_layout.addWidget(self.create_attribute_field(),5)

        # self.frame_layout.setSpacing(1)
        frame.setLayout(self.frame_layout)

        return frame

    def create_list_subtext_frame(self) -> QFrame:
        list_subtext_frame = QFrame()
        self.list_subtext_layout = QVBoxLayout()
        self.list_subtext_layout.setContentsMargins(0, 0, 0, 0)

        list_subtext_label_font = QFont()
        list_subtext_label_font.setItalic(True)

        self.list_subtext_label = QLabel()
        self.list_subtext_label.setText(self.list_subtext_label_text)
        self.list_subtext_label.setFont(list_subtext_label_font)
        self.list_subtext_label.setWordWrap(True)
        self.list_subtext_label.setAlignment(Qt.AlignmentFlag.AlignTop )

        self.list_subtext_layout.addWidget(self.list_subtext_label)
        self.add_extra_elements_to_list_subtext_layout()

        list_subtext_frame.setLayout(self.list_subtext_layout)

        return list_subtext_frame

    def clicked_item_listener(self, table_coord: QModelIndex) -> None:
        """
        Intended to connect to self.list_gui.clicked signal which is triggered when an item in the
        list is clicked

        Responds to item click events in the list widget by checking if the clicked item is a
        folder. If the item is a folder, it toggles the expand state of that item in the GUI.

        :param table_coord: The index of the clicked item in the table.
        :type table_coord: QModelIndex

        :return: None
        """

        table_coord = table_coord.siblingAtColumn(0)
        if self.is_item_a_type_folder_at_row(table_coord):
           self.list_gui.toggle_expand_state_of_item_at_row(table_coord)
        else:
          self.asset_clicked_listener()

    def is_item_a_type_folder_at_row(self, model_index):
        folder_item = self.list_gui.model.itemFromIndex(model_index)
        if folder_item.hasChildren():
            return folder_item
        else:
            return None

    def fill_list(self, source_object: Optional[AIMObject], objects: Collection, last_added) -> None:
        list_label_text = "None"
        if self.list_label:
            list_label_text = self.list_label.text()

        timing_ref = f"fill_list_{list_label_text}"
        object_list_length = len(objects)
        OTLLogger.logger.debug(
            f"Execute AbstractInstanceListWidget.fill_list() ({object_list_length} objects) in list: {list_label_text} ",
            extra={"timing_ref": timing_ref})

        # sourcery skip: remove-dict-keys
        # objects = RelationChangeDomain.objects
        self.list_gui.setSortingEnabled(False)
        self.list_gui.clear()

        item_list = []
        type_to_instance_dict = {}
        self.list_gui.itemDelegate().first_paint = False
        self.list_gui.itemDelegate().second_paint = False

        text_and_data_per_element = self.extract_text_and_data_per_item(source_object, objects,
                                                                        last_added)

        if self.needs_source_object and not source_object:
            self.add_no_asset_selected_placeholder()
        elif not objects:
            self.add_no_options_placeholder()


        for text_and_data in text_and_data_per_element:

            abbr_typeURI = text_and_data['text'].typeURI

            if abbr_typeURI in type_to_instance_dict.keys():
                type_to_instance_dict[abbr_typeURI].append(text_and_data)
            else:
                type_to_instance_dict[abbr_typeURI] = [text_and_data]

        folder_items_expanded = []
        previously_selected_item = None
        self.multi_col_list = False
        for asset_type, text_and_data_list in type_to_instance_dict.items():

            add_folder_based_on_search_text = False
            folder_item = self.create_asset_type_standard_item(asset_type,text_and_data_list)
            folder_item_font = QFont()
            folder_item_font.setBold(True)
            folder_item_font.setPointSize(10)
            folder_item.setFont(folder_item_font)

            # self.type_to_items_dict[asset_type] = []

            if asset_type not in self.type_open_status:
                self.type_open_status[asset_type] = False
            elif self.type_open_status[asset_type]:
                folder_items_expanded.append(folder_item)

            for text_and_data in text_and_data_list:

                instance_item_tuple: tuple[QStandardItem] = self.create_instance_standard_item(text_and_data)
                search_match = False

                # search on all columns
                for instance_item in instance_item_tuple:
                    if (    self.search_text in instance_item.text().lower() or
                            self.search_text in folder_item.text().lower()):
                        search_match = True

                        instance_item.setEditable(False)  # Make the item name non-editable
                        if self.is_last_added(text_and_data):
                            instance_item.setBackground(QBrush(QColor("#ecf0f1")))


                if search_match:
                    if self.is_previously_selected_requirement(text_and_data):
                        previously_selected_item = instance_item_tuple[0]

                    if len(instance_item_tuple) > 1:
                        instance_item_tuple[1].setData(instance_item_tuple[0].data(self.data_1_index), self.data_1_index)
                        instance_item_tuple[1].setData(instance_item_tuple[0].data(self.item_type_data_index), self.item_type_data_index)
                        instance_item_tuple[1].setEditable(False)
                        self.multi_col_list = True

                    folder_item.appendRow(instance_item_tuple)
                    add_folder_based_on_search_text = True

                    if self.search_text:
                        #if you are searching then open all the folders that have the results
                        folder_items_expanded.append(folder_item)

            if add_folder_based_on_search_text:
                item_list.append(folder_item)


        for folder_item in item_list:
            folder_item.sortChildren(0,Qt.SortOrder.AscendingOrder)
            if self.multi_col_list:
                padding_item = QStandardItem("")
                padding_item.setData(folder_item.data(self.data_1_index), self.data_1_index)
                padding_item.setData(folder_item.data(self.item_type_data_index), self.item_type_data_index)
                padding_item.setEditable(False)  # Make the folder name non-editable
                padding_item.setSelectable(False)  # Optional: make the folder itself non-selectable

                self.list_gui.addItem((folder_item, padding_item))
            else:
                self.list_gui.addItem(folder_item)

        self.list_gui.setSortingEnabled(True)
        self.list_gui.sortByColumn(0, Qt.SortOrder.AscendingOrder)

        self.list_gui.expandAll()
        self.list_gui.resizeColumnToContents(0)  # Resizes the first column based on its content
        if self.multi_col_list:
            self.list_gui.resizeColumnToContents(1)

        self.list_gui.collapseAll()

        # expand previously expanded items
        for folder_item in folder_items_expanded:
            folder_item_index = self.list_gui.model.indexFromItem(folder_item)
            self.list_gui.expand(folder_item_index)

        # select previously selected item
        self.select_object_id(previously_selected_item=previously_selected_item)
        if not previously_selected_item:
            self.set_list_button_enabled( False)
            self.selected_object = None
            self.selected_item = None


        if self.needs_source_object and not source_object:
            self.list_gui.model.setHeaderData(0, Qt.Orientation.Horizontal,"")
        elif not objects:
            self.list_gui.model.setHeaderData(0, Qt.Orientation.Horizontal,"")
        else:
            for i in range(len(self.labels)):
                self.list_gui.model.setHeaderData(i, Qt.Orientation.Horizontal, self.labels[i])

        OTLLogger.logger.debug(
            f"Execute AbstractInstanceListWidget.fill_list() ({object_list_length} objects) in list: {list_label_text} ",
            extra={"timing_ref": timing_ref})

    @abc.abstractmethod
    def create_instance_standard_item(self, text_and_data):
        raise NotImplementedError

    @abc.abstractmethod
    def create_button(self):
        raise NotImplementedError

    @abc.abstractmethod
    def extract_text_and_data_per_item(self, source_object, objects, last_added):
        raise NotImplementedError

    @abc.abstractmethod
    def on_item_selectionChange_listener(self, selected: QItemSelectionModel, deselected:QItemSelectionModel):
        raise NotImplementedError

    @abc.abstractmethod
    def is_last_added(self, text_and_data: dict):
        raise NotImplementedError

    def create_attribute_field(self) -> None:

        self.attribute_field.setColumnCount(2)
        self.attribute_field.setProperty('class', 'attribute_field')
        self.attribute_field.setHeaderHidden(True)

        header = self.attribute_field.header()
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
        header.setStretchLastSection(False)

        self.add_attribute_field_placeholder(self.attribute_field,
                                             self.get_no_instance_selected_message())

        return self.attribute_field

    def fill_object_attribute_field(self, object_attribute_dict:dict) -> None:

        self.attribute_field.clear()

        if not object_attribute_dict:
           self.add_attribute_field_placeholder(self.attribute_field,
                                                self.get_no_instance_selected_message())

        for attribute, value in object_attribute_dict.items():
            list_item = QTreeWidgetItem()
            list_item.setText(0, attribute)
            list_item.setText(1, str(value))
            self.attribute_field.addTopLevelItem(list_item)

    def create_search_bar(self) -> QFrame:
        frame = QFrame()
        frame_layout = QHBoxLayout()
        frame_layout.setContentsMargins(5, 5, 0, 5)
        self.search_bar = QLineEdit()
        self.search_bar.setPlaceholderText(self._("search_button"))
        self.search_bar.textChanged.connect(self.search_listener)

        self.clear_search_bar_button = QPushButton()
        self.set_clear_icon(self.clear_search_bar_button)
        self.clear_search_bar_button.clicked.connect(self.clear_search_listener)
        self.clear_search_bar_button.setProperty('class', 'secondary-button')

        frame_layout.addWidget(self.search_bar)
        frame_layout.addWidget(self.clear_search_bar_button)
        frame.setLayout(frame_layout)
        return frame

    def search_listener(self,text:str) -> None:
        self.search_text = text.lower()
        if not self.search_text:
            self.set_all_folder_items_collapsed()

        self.update_this_gui_list_content()


        # RelationChangeDomain.update_frontend()

    def clear_search_listener(self) -> None:
        self.search_bar.setText("")
        RelationChangeDomain.update_frontend()

    def set_search_text(self, text) -> None:
        self.search_bar.setText(text)

    def filter_on_search_text(self, items:list[QListWidgetItem]) -> list[QListWidgetItem]:

        return   [item for item in items
                  if self.search_text in item.text().lower()]

    def get_no_instance_selected_message(self):
        return self._("no_instance_selected")


    def create_asset_type_standard_item(self, asset_type:str, text_and_data_list) -> QStandardItem:
        item_count = len(text_and_data_list)
        selected_item_count = 0

        item = QStandardItem()
        self.set_type_folder_text(type_folder_item= item,
                                  otl_type= asset_type,
                                  item_count=item_count,
                                  selected_item_count=selected_item_count)
        item.setEditable(False)  # Make the folder name non-editable
        item.setSelectable(False)  # Optional: make the folder itself non-selectable

        item.setData(asset_type,self.data_1_index)
        item.setData("type", self.item_type_data_index)
        item.setData(False, self.data_last_added_index)

        # there are only so many data indexes you can use so we have to bundle information
        item.setData([selected_item_count,item_count],self.data_item_count_index)

        return item

    def set_type_folder_text(self, type_folder_item, otl_type, item_count, selected_item_count = 0):
        if not selected_item_count:
            type_folder_item.setText(f"{otl_type} ({item_count})")
        else:
            type_folder_item.setText(f"{otl_type} ({selected_item_count}/{item_count})")

    def update_selected_count_data(self, type_folder_item, selected_item_count):
        item_count = type_folder_item.data(self.data_item_count_index)[1]
        type_folder_item.setData([selected_item_count, item_count], self.data_item_count_index)
        return item_count

    def reset_selected_item_count(self, type_folder_item):
        selected_item_count = 0
        item_count = self.update_selected_count_data(type_folder_item, selected_item_count)
        asset_type = type_folder_item.data(self.data_1_index)
        self.set_type_folder_text(type_folder_item=type_folder_item,
                                  otl_type=asset_type,
                                  item_count=item_count,
                                  selected_item_count=selected_item_count)

    def select_object_id(self, previously_selected_item: QStandardItem):
        pass

    def is_previously_selected_requirement(self, text_and_data):
        return False

    def record_expanse_listener(self, index):
        folder_item: QStandardItem = self.is_item_a_type_folder_at_row(index)
        if folder_item:
            asset_type = folder_item.data(self.data_1_index)
            self.type_open_status[asset_type] = True

    def record_collapse_listener(self, index):

        folder_item: QStandardItem = self.is_item_a_type_folder_at_row(index)
        if folder_item:
            asset_type = folder_item.data(self.data_1_index)
            self.type_open_status[asset_type] = False


    def expand_folder_of(self,typeURI: str):
        self.type_open_status[typeURI] = True

    def set_list_button_enabled(self, item_selected:bool):
        pass

    def set_all_folder_items_collapsed(self):
        self.type_open_status.clear()

    def add_direction_icon_to_item(self, instance_item: QStandardItem, direction: str, typeURI: str):
        direction_icon_path = f'{str(IMG_DIR)}/bidirect.png'
        if direction == "-->":
            direction_icon_path = f'{str(IMG_DIR)}/right.png'
        elif direction == "<--":
            direction_icon_path = f'{str(IMG_DIR)}/left.png'

        pixmap = QPixmap(direction_icon_path)
        self.apply_relation_color(pixmap, typeURI)

        instance_item.setIcon(QIcon(pixmap))

    def add_colored_relation_bol_icon_to_item(self, instance_item: QStandardItem, typeURI: str):
        direction_icon_path = f'{str(IMG_DIR)}/bol.png'

        pixmap = QPixmap(direction_icon_path)
        self.apply_relation_color(pixmap, typeURI)

        instance_item.setIcon(QIcon(pixmap))

    def apply_relation_color(self, pixmap, typeURI):
        painter = QPainter(pixmap)
        painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)
        color_code = '000000'
        if typeURI in self.color_legend.keys():
            color_code = self.color_legend[typeURI]
        color = QColor(f"#{color_code}")  # Choose the color you want
        painter.setBrush(color)
        painter.setPen(color)
        painter.drawRect(pixmap.rect())  # Draw over the pixmap with the chosen color
        painter.end()

    def add_no_asset_selected_placeholder(self):
        pass

    def add_no_options_placeholder(self):
        place_holder_item = QStandardItem(
            self._("no_options_available"))
        place_holder_item.setEditable(False)
        place_holder_item.setEnabled(False)
        place_holder_item.setSelectable(False)

        placeholder_font = QFont()
        placeholder_font.setItalic(True)
        place_holder_item.setFont(placeholder_font)

        # padding_item = QStandardItem("")
        # padding_item.setEditable(False)
        # padding_item.setEnabled(False)
        # padding_item.setSelectable(False)

        self.list_gui.addItem([place_holder_item])

    @classmethod
    def add_attribute_field_placeholder(cls, field,text):
        place_holder_item = QTreeWidgetItem()
        place_holder_item.setText(0, text)

        placeholder_font = QFont()
        placeholder_font.setItalic(True)
        place_holder_item.setFont(0,placeholder_font)
        field.addTopLevelItem(place_holder_item)

    def add_loading_placeholder(self):
        place_holder_item = QStandardItem(
            self._("loading"))
        place_holder_item.setEditable(False)
        place_holder_item.setEnabled(False)
        place_holder_item.setSelectable(False)

        placeholder_font = QFont()
        placeholder_font.setItalic(True)
        place_holder_item.setFont(placeholder_font)

        self.list_gui.addItem(place_holder_item)

    def clear(self):
        self.list_gui.clear()
        self.attribute_field.clear()

    # noinspection PyMethodMayBeStatic
    def set_clear_icon(self, button: QPushButton):
        button.setIcon(qta.icon('mdi.close', color=Styling.button_icon_color))

    def update_color_scheme(self):
        self.set_clear_icon(self.clear_search_bar_button)

    def asset_clicked_listener(self):
        pass

    def get_current_list_content_dict(self) -> dict:
        return self.id_to_object_with_text_and_data_dict

    def add_extra_elements_to_list_subtext_layout(self) -> None:
        pass

    def get_adjustable_subtext_frame(self):
        return self.list_subtext_frame

    @abc.abstractmethod
    def update_this_gui_list_content(self):
        pass


from PyQt6.QtWidgets import QWidget, QMessageBox


class YesOrNoOrAbortNotificationWindow(QMessageBox):

    def __init__(self, message:str, title:str = "notification"):
        self.message = message
        self.title = title
        super().__init__()

        self.setWindowTitle(self.title)
        self.setText(self.message)
        self.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Abort)


from typing import Optional, List

from PyQt6.QtWidgets import QWidget, QFrame
from otlmow_model.OtlmowModel.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject

from otlmow_gui.GUI.screens.Screen import Screen

class RelationChangeScreenInterface(Screen):
    def set_gui_lists_to_loading_state(self) -> None:
        pass

    def paintEvent(self, a0):
        pass
    def synchronize_subtext_label_heights(self) -> None:
        pass
    def init_ui(self) -> None:
        pass

    def create_menu(self) -> QWidget:
        pass

    def fill_object_list(self, objects: List[RelationInteractor]) -> None:
        pass

    def fill_existing_relations_list(self, relations_objects: List[RelatieObject],
                                     last_added: list[RelatieObject] = None) -> None:
        pass

    def fill_possible_relations_list(self, source_object: Optional[RelationInteractor],
                                     relations: dict[str, list[RelatieObject]],
                                     last_added=None) -> None:
        pass

    def horizontal_layout(self) -> QFrame:
        pass


    def fill_object_attribute_field(self, object_attribute_dict: dict) -> None:
        pass

    def fill_possible_relation_attribute_field(self,
                                               possible_relation_attribute_dict: dict) -> None:
        pass
    def fill_existing_relation_attribute_field(self,
                                               existing_relation_attribute_dict: dict) -> None:
        pass

    def expand_existing_relations_folder_of(self, relation_typeURI: str) -> None:
        pass

    def expand_possible_relations_folder_of(self, relation_typeURI: str) -> None:
        pass

    def showMultiSelectionHeeftBetrokkeneAttributeDialogWindow(self,
                                                               data_list_and_relation_objects: list) -> None:
        pass

    def clear_possible_relation_elements(self):
        pass

    def get_current_object_list_content_dict(self) -> dict[str, list]:
        pass

    def set_selected_object(self, identificator: str):
        pass

    def is_show_all_OTL_relations_checked(self) -> bool:
        pass



class ExcelFileUnavailableError(Exception):
    def __init__(self,file_path:str, exception: Exception,*args):
        super().__init__(exception,*args)
        self.error_window_message_key = "permission_to_file_was_denied_likely_due_to_the_file_being_open_in_excel"
        self.error_window_title_key = "permission_denied"
        self.file_path = file_path


from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLabel, QDialogButtonBox

from otlmow_gui.Domain.step_domain.InsertDataDomain import InsertDataDomain
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class RevalidateDocumentsWindow:

    def __init__(self, screen, language_settings):
        self.screen =screen
        self._ = language_settings
        self.init_ui()

    def init_ui(self):
        dialog = QDialog()
        dialog.setModal(True)
        dialog.setWindowTitle(self._("revalidate_documents_files_title"))
        layout = QVBoxLayout()
        question_label = QLabel()
        question_label.setText(self._("revalidate_documents_files_question"))
        layout.addWidget(question_label)
        button_box = self.create_button_box()
        button_box.accepted.connect(lambda: self.validate_documents_again(dialog))
        button_box.rejected.connect(dialog.reject)
        layout.addWidget(button_box)
        dialog.setLayout(layout)
        dialog.show()
        dialog.exec()

    def create_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._('continue'))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))
        return button_box

    def validate_documents_again(self, dialog):
        missing_project_files = InsertDataDomain.check_current_project_project_files_existence()
        if len(missing_project_files):
            self.screen.show_missing_project_files_notification_window(missing_project_files)
        else:
            create_task_reraise_exception(self.screen.validate_documents())
        dialog.close()




from pathlib import Path

from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.enums import FileState


class ProjectFile:
    """
    Represents a project file with its associated file path and state of validation.
    This class initializes a project file object with the specified file path and state,
    ensuring that the file path is stored as a Path object.

    Args:
        file_path (Path | str): The path of the project file, which can be a string or a Path object.
        state (FileState): The state of the project file.

    Attributes:
        file_path (Path): The path of the project file.
        state (FileState): The state of the project file.
    """

    def __init__(self, file_path: Path| str, state: FileState):
        if isinstance(file_path, str):
            file_path = Path(file_path)

        self.file_path: Path = file_path
        self.state: FileState = state

    def __eq__(self, __value):
        return Helpers.equal_paths(self.file_path,__value.file_path) and self.state == __value.state




from abc import abstractmethod
from pathlib import Path

from PyQt6.QtWidgets import QFrame, QWidget

from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.GUI.screens.Screen import Screen


class TemplateScreenInterface(Screen):

    def __init__(self):
        super().__init__()


    @abstractmethod
    def init_ui(self) -> None:
        pass


    @abstractmethod
    def options_menu(self) -> QFrame:
        pass


    @abstractmethod
    def create_example_generation_container(self) -> QFrame:
        pass


    @abstractmethod
    def template_menu(self) -> QWidget:
        pass


    @abstractmethod
    def subset_title_and_button(self)  -> QFrame:
        pass


    @abstractmethod
    def create_operator_info_field(self) -> QFrame:
        pass


    @abstractmethod
    def create_otl_version_field(self) -> QFrame:
        pass


    @abstractmethod
    def create_subset_name_field(self) -> QFrame:
        pass


    @abstractmethod
    def subset_info_list(self) -> QFrame:
        pass


    @abstractmethod
    def create_list(self) -> QFrame:
        pass


    @abstractmethod
    def update_project_info(self, project:Project) -> None:
        pass



    @abstractmethod
    def select_all_classes_clicked(self) -> None:
        pass

    @abstractmethod
    def set_gui_list_to_loading_state(self) -> None:
        pass

    @abstractmethod
    def set_gui_list_to_no_classes_found(self) -> None:
        pass

    @abstractmethod
    def set_classes(self, classes,selected_classes: list[int],
                    all_classes_selected_checked:bool, has_a_class_with_deprecated_attributes) -> None:
        pass

    def open_folder_of_created_template(self, document_path:Path):
        pass

    @abstractmethod
    def update_label_under_list(self,total_amount_of_items:int,counter:int) -> None:
        pass

    @abstractmethod
    def set_all_classes_selected(self) -> None:
        pass

    @abstractmethod
    def deselect_all_classes(self) -> None:
        pass

    @abstractmethod
    def update_all_classes_selected(self, state: bool) -> None:
        pass


from collections import namedtuple

from PyQt6.QtCore import Qt, QItemSelectionModel
from PyQt6.QtGui import QStandardItem
from PyQt6.QtWidgets import QFrame
from otlmow_model.OtlmowModel.Helpers import OTLObjectHelper

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.screens.RelationChange_elements.AbstractInstanceListWidget import \
    AbstractInstanceListWidget

from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class ExistingRelationListWidget(AbstractInstanceListWidget):

    Text = namedtuple('text', ['typeURI', 'name_source', 'direction', 'name_target','full_typeURI'])
    Data = namedtuple('data', ["index", "last_added"])

    def __init__(self, language_settings,parent):
        self._ = language_settings
        labels = [self._("source_asset"),self._("target_asset")]
        super().__init__(language_settings=language_settings,parent=parent,labels=labels)

        self.list_label_text = self._("existing_relations_list")
        self.list_subtext_label_text = self._("existing_relations_description")
        self.attribute_field_label_text = self._("existing_relation_attributes")

    def create_object_list_gui(self, multi_select: bool = False) -> QFrame:
        frame = super().create_object_list_gui(multi_select)
        self.frame_layout.setContentsMargins(11, 0, 0, 0)
        return frame

    def on_item_selectionChange_listener(self, selected: QItemSelectionModel, deselected:QItemSelectionModel):
        no_item_selected = True

        for index in deselected.indexes():
            if index.column() == 0:
                item = self.list_gui.model.itemFromIndex(index)
                if item and item.isSelectable():
                    type_folder_item = item.parent()
                    self.reset_selected_item_count(type_folder_item=type_folder_item)

        dict_type_to_type_folder_item = {}
        dict_type_to_selected_item_count = {}
        # Get the currently selected indexes
        for index in self.list_gui.selectionModel().selectedIndexes():
            if index.column() == 0:
                item = self.list_gui.model.itemFromIndex(index)
                if item and item.isSelectable():
                    no_item_selected = False

                    # keep count of selected items in folder
                    parent_type_folder_item = item.parent()
                    parent_type_folder_type = parent_type_folder_item.data(self.data_1_index)
                    if parent_type_folder_type in dict_type_to_selected_item_count.keys():
                        dict_type_to_selected_item_count[parent_type_folder_type] += 1
                    else:
                        dict_type_to_selected_item_count[parent_type_folder_type] = 1
                        dict_type_to_type_folder_item[
                            parent_type_folder_type] = parent_type_folder_item

        # update the selected_counts on all type_folder_items
        for type_folder_type, selected_item_count in dict_type_to_selected_item_count.items():
            OTLLogger.logger.debug(f"{type_folder_type}: {selected_item_count}")

            type_folder_item = dict_type_to_type_folder_item[type_folder_type]

            item_count = self.update_selected_count_data(type_folder_item=type_folder_item,
                                                         selected_item_count=selected_item_count)

            # apply new information to the folder_item display text
            self.set_type_folder_text(type_folder_item=type_folder_item,
                                      otl_type=type_folder_type,
                                      item_count=item_count,
                                      selected_item_count=selected_item_count)

        self.existing_relations_selected()
        self.set_list_button_enabled(not no_item_selected)

    def set_list_button_enabled(self, item_selected:bool):
        if not item_selected:
            if self.list_button.isEnabled():
                self.list_button.setEnabled(False)
        elif not self.list_button.isEnabled():
            self.list_button.setEnabled(True)

    def create_button(self):
        self.list_button.setText(self._('remove_relation'))
        self.list_button.setDisabled(True)
        self.list_button.clicked.connect(
            self.remove_existing_relations_listener)
        self.list_button.setProperty('class', 'remove_relation-button')
        return self.list_button

    def create_instance_standard_item(self, text_and_data):
        text = f"{text_and_data['text'].name_source}"
        instance_item = QStandardItem(text)
        instance_item.setData(text_and_data["data"].index, self.data_1_index)
        instance_item.setData("instance", self.item_type_data_index)
        instance_item.setData(text_and_data["data"].last_added, self.data_last_added_index)

        instance_item.setTextAlignment(Qt.AlignmentFlag.AlignRight)

        text2 = f"{text_and_data['text'].name_target}"
        instance_item2 = QStandardItem(text2)
        instance_item2.setTextAlignment(Qt.AlignmentFlag.AlignTop)
        # instance_item2.setData(text_and_data["data"].index, self.data_1_index)
        # instance_item2.setData("instance", self.item_type_data_index)

        self.add_direction_icon_to_item(instance_item=instance_item2,
                                        direction=text_and_data['text'].direction,
                                        typeURI= text_and_data['text'].full_typeURI)

        return instance_item, instance_item2

    def existing_relations_selected(self):
        indices: list[int] = self.get_selected_data()
        create_task_reraise_exception(
        RelationChangeDomain.select_existing_relation_indices(indices))

    def get_selected_data(self):
        return [
            self.list_gui.model.itemFromIndex(model_i).data(self.data_1_index)
            for model_i in self.list_gui.selectionModel().selectedIndexes() if model_i.column() == 0]

    def remove_existing_relations_listener(self):
        indices: list[int] = sorted(self.get_selected_data(), reverse=True)

        # for index in indices:
        #     RelationChangeDomain.remove_existing_relation(index)
        create_task_reraise_exception(RelationChangeDomain.remove_multiple_existing_relations(indices))


    def extract_text_and_data_per_item(self, source_object, objects, last_added):
        list_of_corresponding_values = []
        Text = self.Text
        Data = self.Data

        for i, relation_object in enumerate(objects):

            source_object = RelationChangeDomain.get_object(
                identificator=relation_object.bronAssetId.identificator)
            target_object = RelationChangeDomain.get_object(
                identificator=relation_object.doelAssetId.identificator)
            screen_name_source = RelationChangeHelpers.get_screen_name(otl_object=source_object)
            screen_name_target = RelationChangeHelpers.get_screen_name(otl_object=target_object)
            add_namespace = RelationChangeHelpers.is_unique_across_namespaces(
                typeURI=relation_object.typeURI,
                objects=RelationChangeDomain.shown_objects)
            abbr_typeURI = RelationChangeHelpers.get_abbreviated_typeURI(
                typeURI=relation_object.typeURI,
                add_namespace=add_namespace,
                is_relation=OTLObjectHelper.is_relation(relation_object))

            direction = RelationChangeHelpers.get_screen_icon_direction("Unspecified")

            if OTLObjectHelper.is_directional_relation(relation_object):
                direction = RelationChangeHelpers.get_screen_icon_direction(
                    "Source -> Destination")

            list_of_corresponding_values.append({
                "text": Text(abbr_typeURI, screen_name_source, direction, screen_name_target,relation_object.typeURI),
                "data": Data(i,relation_object in last_added)
            })

        list_of_corresponding_values.sort(key=lambda val: (
            val['text'].typeURI if val['text'].typeURI is not None else "xxx",
            val['text'].name_source if val['text'].name_source is not None else "xxx",
            val['text'].name_target if val['text'].name_target is not None else "xxx"))

        return list_of_corresponding_values

    def is_last_added(self, text_and_data: dict):
        return text_and_data["data"].last_added

    def create_asset_type_standard_item(self, asset_type, text_and_data_list):
        folder_item = super().create_asset_type_standard_item(asset_type, text_and_data_list)

        if text_and_data_list:
            full_typeURI = text_and_data_list[0]["text"].full_typeURI
            self.add_colored_relation_bol_icon_to_item(folder_item, full_typeURI)
        return folder_item

    def get_no_instance_selected_message(self):
        return self._("no_relation_selected")

    def update_this_gui_list_content(self):
        RelationChangeDomain.update_frontend_existing_relations()




from PyQt6.QtWidgets import QListWidget

from otlmow_gui.Domain.step_domain.TemplateDomain import TemplateDomain


class ClassListWidget(QListWidget):



    def mouseReleaseEvent(self, e):
        super().mouseReleaseEvent(e)
        self.sync_selected_items_with_backend()

    def sync_selected_items_with_backend(self):
        TemplateDomain.set_selected_indexes([item.data(1)[1] for item in self.selectedItems()])



import qtawesome as qta

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QWidget, QHBoxLayout, QFrame

from otlmow_gui.Domain import global_vars
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget


class StepperWidget(QWidget):
    def __init__(self, _, step: int):
        super().__init__()
        self._ = _
        self.main_window = None
        self.step1 = ButtonWidget()
        self.step2 = ButtonWidget()
        self.step3 = ButtonWidget()
        self.step4 = ButtonWidget()
        self.step_nr = step

    def stepper_widget(self):
        stepper_widget = QWidget()
        lines = [QFrame() for _ in range(3)]
        for line in lines:
            line.setFrameShape(QFrame.Shape.HLine)
            line.setProperty('class', 'stepper-line')

        # sets the text for the stepper buttons and applies classes which hold the style
        steps = [self.step1, self.step2, self.step3, self.step4]
        self.populate_steps(steps)
        layout = self.fill_up_layout(lines)
        stepper_widget.setLayout(layout)
        return stepper_widget

    def populate_steps(self, steps):
        for i, step in enumerate(steps, start=1):
            step.setText(self._(f"step{i}"))
            step.setProperty('class', 'stepper-button')
            color = "#B35F35" if self.step_nr == i else "grey"
            step.setIcon(qta.icon(f'mdi.numeric-{i}-circle', color=color))
            step.clicked.connect(lambda _, index=i: self.main_window.setCurrentIndex(index))

    def fill_up_layout(self, lines):
        horizontal_layout = QHBoxLayout()
        self.enable_steps()
        horizontal_layout.addWidget(self.step1, alignment=Qt.AlignmentFlag.AlignCenter)
        horizontal_layout.addWidget(lines[0])
        horizontal_layout.addWidget(self.step2, alignment=Qt.AlignmentFlag.AlignLeft)
        horizontal_layout.addWidget(lines[1])
        horizontal_layout.addWidget(self.step3, alignment=Qt.AlignmentFlag.AlignLeft)
        horizontal_layout.addWidget(lines[2])
        horizontal_layout.addWidget(self.step4, alignment=Qt.AlignmentFlag.AlignLeft)
        horizontal_layout.addSpacing(50)
        horizontal_layout.setContentsMargins(10, 0, 0, 0)
        return horizontal_layout

    def enable_steps(self):
        if (global_vars.current_project and
                global_vars.current_project.are_all_project_files_in_memory_valid()):
            self.step3.setDisabled(False)
            self.step4.setDisabled(False)
        else:
            # OTLLogger.logger.debug('set disabled')
            self.step3.setDisabled(True)
            self.step4.setDisabled(True)

    def reset_ui(self, _):
        self._ = _
        self.step1.setText(self._("step1"))
        self.step2.setText(self._("step2"))
        self.step3.setText(self._("step3"))
        self.step4.setText(self._("step4"))
        self.enable_steps()



from typing import Callable

from PyQt6.QtWidgets import QMessageBox, QPushButton

from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain


class RemoveProjectWindow(QMessageBox):

    def __init__(self, language_settings: Callable, project: Project):
        super().__init__()
        self._ = language_settings


        self.setWindowTitle(self._("delete"))
        self.setText(self._("project_deletion_question"))
        self.setIcon(QMessageBox.Icon.Warning)
        self.addButton(QPushButton(self._("yes")), QMessageBox.ButtonRole.YesRole)
        self.addButton(QPushButton(self._("no")), QMessageBox.ButtonRole.NoRole)
        self.exec()
        reply = self.buttonRole(self.clickedButton())
        if reply == QMessageBox.ButtonRole.YesRole:
            HomeDomain.remove_project(project)


from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.util.Helpers import Helpers


class VisualisationStateTracker:

    def __init__(self):
        super().__init__()

        self.relations_to_be_inserted: list[RelatieObject] =  []
        self.relations_to_be_removed: list[RelatieObject] = []
        self.clear_all: bool = False

    def insert_relation(self,relation_object:RelatieObject) -> None:
        indexes_list = []
        for i,relation_to_remove in enumerate(self.relations_to_be_removed):
            if Helpers.compare_RelatieObjects(relation_to_remove,relation_object):
                indexes_list.append(i)

        if indexes_list:
            OTLLogger.logger.debug("cancel insert_relation " + str(indexes_list))
            for i in indexes_list:
                # if this relation has been removed before updating the visual
                self.relations_to_be_removed.pop(i)
        else:
            self.relations_to_be_inserted.append(relation_object)

    def remove_relation(self, relation_object: RelatieObject) -> None:

        indexes_list = []
        for i, relation_to_add in enumerate(self.relations_to_be_inserted):
            if Helpers.compare_RelatieObjects(relation_to_add, relation_object):
                indexes_list.append(i)

        if indexes_list:
            OTLLogger.logger.debug("cancel remove_relation " + str(indexes_list))
            for i in indexes_list:
                # if this relation has been added before updating the visual
                self.relations_to_be_inserted.pop(i)
        else:
            self.relations_to_be_removed.append(relation_object)

    def reset_relations_uptodate(self) -> None:
        self.relations_to_be_inserted.clear()
        self.relations_to_be_removed.clear()

    def reset_full_state(self) -> None:
        self.reset_relations_uptodate()
        self.set_clear_all(False)

    def is_uptodate(self) -> bool:
        return (len(self.relations_to_be_inserted) == 0 and
                len(self.relations_to_be_removed) == 0 and
                not self.clear_all)

    def get_clear_all(self) -> bool:
        return self.clear_all

    def set_clear_all(self,clear_all:bool) -> None:
        self.clear_all = clear_all

    def get_to_be_inserted_relations(self) -> list[RelatieObject]:
        OTLLogger.logger.debug("self.relations_to_be_inserted " + str(self.relations_to_be_inserted))
        return self.relations_to_be_inserted

    def get_to_be_removed_relations(self) -> list[RelatieObject]:
        OTLLogger.logger.debug(
            "self.relations_to_be_removed " + str(self.relations_to_be_removed))
        return self.relations_to_be_removed